/*! ====================================================================================================================
 * \file
		v2v_dec.c
 *  \brief
		Copyright information.
 *  \par Copyright statements
		HEVC (JCTVC cfp)

		This software, including its corresponding source code, object code and executable, may only be used for 
		(1) research purposes or (2) for evaluation for standardisation purposes within the joint collaborative team on 
		video coding for HEVC , provided that this copyright notice and this corresponding notice appear in all copies, 
		and that the name of Research in Motion Limited not be used in advertising or publicity without specific, written 
		prior permission.  This software, as defined above, is provided as a proof-of-concept and for demonstration 
		purposes only; there is no representation about the suitability of this software, as defined above, for any purpose.  
 
		THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
		INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
		DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
		SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR 
		SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
		WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
		USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
		
		TRADEMARKS, Product and company names mentioned herein may be the trademarks of their respective owners.  
		Any rights not expressly granted herein are reserved.
 
		Copyright (C) 2010 by Research in Motion Limited, Canada      
		All rights reserved.
 
 *  \par Full Contact Information
		Standards & Licensing Department			(standards-ipr@rim.com)
		Research in Motion
		122 West John Carpenter Parkway
		Irving, TX 75039, USA

 * ====================================================================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>


#include "global.h"
#include "v2v_dec.h"
#include "trees67k.h"
#include "vlcdec.h"


#define Dbuffer         (dep->Dbuffer)
#define Dbits_to_go     (dep->Dbits_to_go)
#define Dbits_used      (dep->Dbits_used)
#define Dcodestrm       (dep->Dcodestrm)
#define Dcodestrm_len   (dep->Dcodestrm_len)


uchar legacy_cabac;
uchar temp_space[TEMP_SIZE];
uint seq_coded_len[TREE_NUM];
uint seq_coded_offset[MAX_CPUS];
uint offset[TREE_NUM];
uint term_offset[TREE_NUM];
uint bit_pos[TREE_NUM];

#define get_byte(){                                         \
                    Dbuffer = Dcodestrm[(*Dcodestrm_len)++];\
                    Dbits_to_go = 7;                        \
                  }

#define vlc_get_byte() {                                     \
    Dbuffer |= Dcodestrm[(*Dcodestrm_len)++] << Dbits_used;  \
    Dbits_used += 8;                                         \
}


uint get_pref_code(DecodingEnvironmentPtr dep) {

    uint n;
    assert(Dbits_to_go == 0);

    get_byte(); Dbits_to_go = 0;
    n = Dbuffer;
    if (!(n & 1)) return n >> 1;
    n >>= 1;

    get_byte(); Dbits_to_go = 0;
    n |= Dbuffer << 7;
    if (!(n & 1)) return 128 + (n >> 1);
    n >>= 1;

    get_byte(); Dbits_to_go = 0;
    n |= Dbuffer << 14;
    if (!(n & 1)) return 16512 + (n >> 1);
    n >>= 1;

    get_byte(); Dbits_to_go = 0;
    n |= Dbuffer << 21;
    return 2113664 + n;
}

void v2v_decode_seq(DecodingEnvironmentPtr dep, int state, int *offset, int scl) {
    uint preflen = 0, phrase = 0;
    uint bitbuf = 0, bitbufsize = 0;

    assert(Dbuffer == 0 && Dbits_used == 0);
    while (1) {
        uint bit = 0;

        if (preflen > 0) {
            --preflen;
        }
        else if (phrase > 1) {
            if ((phrase & 1) == 0) bit = !bit;
            phrase >>= 1;
        }
        else {

            int pos = 0;
            uint32 dpp = dec_tree_complete[state][0];

            while (!(dpp >> 30)) {
                int len = dpp >> 16;
                while (scl && Dbits_used < len) {
                    vlc_get_byte();
                    --scl;
                }
                if (!scl && Dbits_used < len) {
                    if (bitbufsize) {
                        temp_space[(*offset)++] = (uchar)bitbuf;
                    }
                    return;
                }

                pos = (dpp & 0xffff) + (Dbuffer & ((1 << len) - 1));
                Dbuffer >>= len; Dbits_used -= len;
                dpp = dec_tree_complete[state][pos];
            }

            switch (dpp >> 30) {
                case 1:
                    phrase = (1 << ((dpp >> 25) & 0x1f)) | (dpp & 0x1ffffff);
                    if ((phrase & 1) == 0) bit = !bit;
                    phrase >>= 1;    /* Consume next output bit */
                    break;
                case 2:
                    preflen = (dpp & 0xffffff) - 1;
                    phrase = 2;
                    --preflen;    /* Consume next output bit */
                    break;
                case 3:
                    preflen = (dpp & 0xffffff) - 1;
                    phrase = 3;
                    --preflen;    /* Consume next output bit */
                    break;
                /*default:
                    assert(0); */
            }
        }

        //* REMOVE */ fprintf(stderr, "%u\n", bit);
        bitbuf |= bit << bitbufsize;
        if (++bitbufsize == 8) {
            temp_space[(*offset)++] = (uchar)bitbuf;
            bitbuf = bitbufsize = 0;
        }
    }
}


void vlc_decode_seq(DecodingEnvironmentPtr dep, int state, int *offset, int scl) {

    assert(Dbuffer == 0 && Dbits_used == 0);
    while (1) {

        uint32 dpp, pos;
        dpp = vlc_dec_tree[state][pos = 0];

        while (!(dpp >> 30)) {
            int len = dpp >> 16;
            while (scl && Dbits_used < len) {
                vlc_get_byte();
                --scl;
            }
            if (!scl && Dbits_used < len) return;

            pos = (dpp & 0xffff) + (Dbuffer & ((1 << len) - 1));
            Dbuffer >>= len; Dbits_used -= len;
            dpp = vlc_dec_tree[state][pos];
        }

        temp_space[(*offset)++] = (uchar)dpp;
    }
}


