/*! ====================================================================================================================
 * \file
		softq.h
 *  \brief
		Copyright information.
 *  \par Copyright statements
		HEVC (JCTVC cfp)

		This software, including its corresponding source code, object code and executable, may only be used for 
		(1) research purposes or (2) for evaluation for standardisation purposes within the joint collaborative team on 
		video coding for HEVC , provided that this copyright notice and this corresponding notice appear in all copies, 
		and that the name of Research in Motion Limited not be used in advertising or publicity without specific, written 
		prior permission.  This software, as defined above, is provided as a proof-of-concept and for demonstration 
		purposes only; there is no representation about the suitability of this software, as defined above, for any purpose.  
 
		THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
		INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
		DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
		SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR 
		SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
		WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
		USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
		
		TRADEMARKS, Product and company names mentioned herein may be the trademarks of their respective owners.  
		Any rights not expressly granted herein are reserved.
 
		Copyright (C) 2010 by Research in Motion Limited, Canada      
		All rights reserved.
 
 *  \par Full Contact Information
		Standards & Licensing Department			(standards-ipr@rim.com)
		Research in Motion
		122 West John Carpenter Parkway
		Irving, TX 75039, USA

 * ====================================================================================================================
*/

#ifndef _SOFTQ_H_
#define _SOFTQ_H_

#include <stdio.h>
#include "global.h"

#define MAX_NO_TRANSITION  64

#if 0
#define SOFTQ_RECORD_DISTORTION
#endif

typedef enum
{
  //eE0_G0_Eq0,
  eE0_G0_Eq1,
  eE0_G0_Gt1,
  eE0_G0_Gt1_F,
  //
  eE1_G0_Eq0,
  eE1_G0_Eq1,
  eE1_G0_Gt1,
  eE1_G0_Gt1_F,
  //
  eE2_G0_Eq0,
  eE2_G0_Eq1,
  eE2_G0_Gt1,
  eE2_G0_Gt1_F,
  //
  eE3_G0_Eq0,
  eE3_G0_Eq1,
  eE3_G0_Gt1,
  eE3_G0_Gt1_F,
  //
  eEx_G0_Eq0,
  eEx_G0_Eq1,
  eEx_G0_Gt1,
  eEx_G0_Gt1_F,
  //
  eEx_G1_Eq0,
  eEx_G1_Eq1,
  eEx_G1_Gt1,
  eEx_G1_Gt1_F,
  //
  eEx_G2_Eq0,
  eEx_G2_Eq1,
  eEx_G2_Gt1,
  eEx_G2_Gt1_F,
  //
  eEx_G3_Eq0,
  eEx_G3_Eq1,
  eEx_G3_Gt1,
  eEx_G3_Gt1_F,
  //
  eEx_Gx_Eq0,
  eEx_Gx_Eq1,
  eEx_Gx_Gt1,
  eEx_Gx_Gt1_F,
  eHead,
  eEnd,
  eNumOfStates 
} tStateName;

#define EQUAL_ZERO(X) ((X)==eE1_G0_Eq0 || (X)==eE2_G0_Eq0 || (X)==eE3_G0_Eq0 || (X)== eEx_G0_Eq0 || \
					   (X)==eEx_G1_Eq0 || (X)==eEx_G2_Eq0 || (X)==eEx_G3_Eq0 || (X)==eEx_Gx_Eq0)

#define EQUAL_ONE(X)  ((X)==eE0_G0_Eq1 || (X)==eE1_G0_Eq1 || (X)==eE2_G0_Eq1 || (X)==eE3_G0_Eq1 || \
					   (X)==eEx_G0_Eq1 || (X)==eEx_G1_Eq1 || (X)==eEx_G2_Eq1 || (X)==eEx_G3_Eq1 || \
					   (X)==eEx_Gx_Eq1)

#define GT_ONE(X)	  ((X)==eE0_G0_Gt1 || (X)==eE1_G0_Gt1 || (X)==eE2_G0_Gt1 || (X)== eE3_G0_Gt1 || \
					   (X)==eEx_G0_Gt1 || (X)==eEx_G1_Gt1 || (X)==eEx_G2_Gt1 || (X)== eEx_G3_Gt1 || \
					   (X)==eEx_Gx_Gt1)

#define GT_ONE_FF(X)  ((X)==eE0_G0_Gt1_F || (X)==eE1_G0_Gt1_F || (X)==eE2_G0_Gt1_F || (X)==eE3_G0_Gt1_F || \
					   (X)==eEx_G0_Gt1_F || (X)==eEx_G1_Gt1_F || (X)==eEx_G2_Gt1_F || (X)==eEx_G3_Gt1_F || \
					   (X)==eEx_Gx_Gt1_F)

#define COEFF_STATES	(eNumOfStates - 2)

typedef struct tagTargetState
{
  int len; // number of outgoing states for
  tStateName ToState[4];
} tTargetState;

typedef struct tagTransition
{
  int InCoefIndex;
  tStateName InStateName;
  double fcost;
#ifdef SOFTQ_RECORD_DISTORTION
  double distortion;
#endif
} tTransition;


typedef struct tagState
{
  tStateName  name;
  tTransition Tran[MAX_NO_TRANSITION];
  int numT;			//Number of valid transitions
  int minTindex;    //Index for the transition with the minimal cost
  double fminTcost; //the minimal cost
#ifdef SOFTQ_RECORD_DISTORTION
  double fminDistortion;
#endif
  Boolean bAlive;   //Some states may be valid, but not alive. Only states with incoming trans is set as Alive.
} tState;


typedef struct tagEndState
{
  tStateName  name;
  tTransition Tran[MAX_NO_TRANSITION];
  int numT;			//Number of valid transitions
  double fminTcost; //Index for the transition with the minimal cost
#ifdef SOFTQ_RECORD_DISTORTION
  double fminDistortion;
#endif
  int minTindex;
} tEndState ;


typedef struct tagCoeff
{
  tState S[COEFF_STATES];
} tCoeff;

typedef struct
{
  int sign;    // ALWAYS TRUE: 1 > 0, -1 < 0, 0 == 0
  int num_levels;
  int level_low;
  int level_high;
  double dist_low;
  double dist_high;
  double dist_zero;
} qCoeff;

void get_CBP_bits_CABAC(int cbp, int b8, int* bits0, int* bits1);
void init_context(int num_coeff, int type);

void init_sdq_data(int* scaled_coeff, double* efactor, int num_coeff, int q_bits);
int len_binary(BiContextTypePtr ptr_ctx, int symbol);
int len_cabac(tStateName m, int i, int u, int type);
void update_ctx_binary(BiContextTypePtr ptr_ctx, int symbol);

void set_valid_states(int i, Boolean* bValid, int len);
void coeff_to_end(tCoeff * C, int num_coeff, int i, tEndState * End, double lambda, double quant_err[64][COEFF_STATES], int coeff_rate[64][COEFF_STATES][2]);
void update_coeff(int num_coeff, tCoeff *C, int i, double DiscardThr, int quant_out[64][COEFF_STATES]);
void update_state(int num_coeff, int i, tState *pS , double DiscardThr, int quant_out[64][COEFF_STATES]);
void init_trans(tCoeff * C, int num_coeff);

void update_context(int num_coeff, int* y, int type);
void update_context_block(int num_coeff, int* y, int b8, int bcbp_ctx, int type);
void trace_back(tCoeff *C, int num_coeff, tEndState *End, int *y, int quant_out[64][COEFF_STATES]);
void connect_head(tCoeff * C, double* discard_coef_err, int num_coeff, double lamdba, int cbp1,
				  Boolean state_valid[64][COEFF_STATES], int type);
void head_to_coeff(tCoeff * C, double* discard_coef_err, int num_coeff, int i, double lamdba, int cbp1,
				   Boolean state_valid[COEFF_STATES], int type);
void coeff_trans(tCoeff *C, int iC, int jC, double lambda, Boolean state_valid[64][COEFF_STATES],
							int quant_out[64][COEFF_STATES], double quant_err[64][COEFF_STATES], int coeff_rate[64][COEFF_STATES][2], int type);
void state_trans(tCoeff *C, int iC, tStateName snfrom, int jC, tStateName snto,
				   double lambda, int quant_out[64][COEFF_STATES], double quant_err[64][COEFF_STATES],
					   int coeff_rate[64][COEFF_STATES][2], int type);
void init_quant(int* scaled_coeff, double* ef, int num_coeff, int qbits, Boolean state_valid[64][COEFF_STATES],
				int quant_out[64][COEFF_STATES], double quant_err[64][COEFF_STATES],
				int coeff_rate[64][COEFF_STATES][2], int type);
void sdq_cabac(int* sign_coeff, int* Idx, int* scaled_coeff, double* efactor, int len, int qp_per, int* y,
			   double lambda, double* RDcost, int *num_coeff, int init_sdq_cabac_context, int type);

void sdq_cabac8x8
(int* sign_coeff, int* Idx, int* scaled_coeff, double* efactor, int len, int qp_per, int* y,
			   double lambda, double* RDcost, int *num_coeff, int init_sdq_cabac_context, int type, int b8);

#endif
