/*! ====================================================================================================================
 * \file
		softq.c
 *  \brief
		Copyright information.
 *  \par Copyright statements
		HEVC (JCTVC cfp)

		This software, including its corresponding source code, object code and executable, may only be used for 
		(1) research purposes or (2) for evaluation for standardisation purposes within the joint collaborative team on 
		video coding for HEVC , provided that this copyright notice and this corresponding notice appear in all copies, 
		and that the name of Research in Motion Limited not be used in advertising or publicity without specific, written 
		prior permission.  This software, as defined above, is provided as a proof-of-concept and for demonstration 
		purposes only; there is no representation about the suitability of this software, as defined above, for any purpose.  
 
		THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
		INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
		DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
		SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR 
		SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
		WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
		USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
		
		TRADEMARKS, Product and company names mentioned herein may be the trademarks of their respective owners.  
		Any rights not expressly granted herein are reserved.
 
		Copyright (C) 2010 by Research in Motion Limited, Canada      
		All rights reserved.
 
 *  \par Full Contact Information
		Standards & Licensing Department			(standards-ipr@rim.com)
		Research in Motion
		122 West John Carpenter Parkway
		Irving, TX 75039, USA

 * ====================================================================================================================
*/

#include "math.h"
#include "stdlib.h"
#include "malloc.h"
#include "memory.h"

#include "global.h"

#include "biariencode.h"
#include "mb_access.h"
#include "mbuffer.h"

#include "cabac.h"	// init_context()
#include "softq.h"


#define SIGNBIT		32768
#define MAXCOST		1073741824

double soft_rd_cost;

extern const int estErr4x4[6][4][4];
extern const byte rLPS_table_64x4[64][4];
extern const unsigned short AC_next_state_MPS_64[64];
extern const unsigned short AC_next_state_LPS_64[64];
extern const byte renorm_table_32[32];
extern int *qp_per_matrix;
extern int *qp_rem_matrix;
extern StorablePicture *enc_picture;
extern const int entropyBits[128];

// in cabac.c
extern const int type2ctx_bcbp[];
extern const int maxpos[];
extern const int type2ctx_map []; 
extern const int type2ctx_last[]; 
extern const int type2ctx_one []; 
extern const int type2ctx_abs []; 
extern const int max_c2       []; 
extern const int* pos2ctx_map [];
extern const int* pos2ctx_last[];


// coefficients
static tCoeff C[256];
static qCoeff q_coeff[256];
static TextureInfoContexts Init_ctx;
static TextureInfoContexts Local_ctx[64];

//from NumEq1_NumGt1 to the Pin1,Pin2 contex models
static int nameCtxPin1[eNumOfStates] = 
{1,1,1, 2,2,2,2, 3,3,3,3, 4,4,4,4, 4,4,4,4, 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,  0,0};
static int nameCtxPin2[eNumOfStates] = 
{0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0, 1,1,1,1, 2,2,2,2, 3,3,3,3, 4,4,4,4,  0,0};

//////////////////////////////////////////////////////
// outgoing states of all states
static tTargetState TargetState[COEFF_STATES] =
{   
  {4, {eE1_G0_Eq0,  eE1_G0_Eq1,  eE1_G0_Gt1,  eE1_G0_Gt1_F}},  //eE0_G0_Eq1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE0_G0_Gt1, 
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE0_G0_Gt1_F, 
  //            
  {4, {eE1_G0_Eq0,  eE1_G0_Eq1,  eE1_G0_Gt1,  eE1_G0_Gt1_F}},  //eE1_G0_Eq0,   
  {4, {eE2_G0_Eq0,  eE2_G0_Eq1,  eE2_G0_Gt1,  eE2_G0_Gt1_F}},  //eE1_G0_Eq1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE1_G0_Gt1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE1_G0_Gt1_F, 
  //
  {4, {eE2_G0_Eq0,  eE2_G0_Eq1,  eE2_G0_Gt1,  eE2_G0_Gt1_F}},  //eE2_G0_Eq0,   
  {4, {eE3_G0_Eq0,  eE3_G0_Eq1,  eE3_G0_Gt1,  eE3_G0_Gt1_F}},  //eE2_G0_Eq1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE2_G0_Gt1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE2_G0_Gt1_F, 
  //     
  {4, {eE3_G0_Eq0,  eE3_G0_Eq1,  eE3_G0_Gt1,  eE3_G0_Gt1_F}},  //eE3_G0_Eq0,   
  {4, {eEx_G0_Eq0,  eEx_G0_Eq1,  eEx_G0_Gt1,  eEx_G0_Gt1_F}},  //eE3_G0_Eq1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE3_G0_Gt1,
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eE3_G0_Gt1_F, 
  //       
  {4, {eEx_G0_Eq0,  eEx_G0_Eq1,  eEx_G0_Gt1,  eEx_G0_Gt1_F}},  //eEx_G0_Eq0,   
  {4, {eEx_G0_Eq0,  eEx_G0_Eq1,  eEx_G0_Gt1,  eEx_G0_Gt1_F}},  //eEx_G0_Eq1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eEx_G0_Gt1,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eEx_G0_Gt1_F, 
  //
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eEx_G1_Eq0,   
  {4, {eEx_G1_Eq0,  eEx_G1_Eq1,  eEx_G1_Gt1,  eEx_G1_Gt1_F}},  //eEx_G1_Eq1,   
  {4, {eEx_G2_Eq0,  eEx_G2_Eq1,  eEx_G2_Gt1,  eEx_G2_Gt1_F}},  //eEx_G1_Gt1,   
  {4, {eEx_G2_Eq0,  eEx_G2_Eq1,  eEx_G2_Gt1,  eEx_G2_Gt1_F}},  //eEx_G1_Gt1_F,   
  //
  {4, {eEx_G2_Eq0,  eEx_G2_Eq1,  eEx_G2_Gt1,  eEx_G2_Gt1_F}},  //eEx_G2_Eq0,   
  {4, {eEx_G2_Eq0,  eEx_G2_Eq1,  eEx_G2_Gt1,  eEx_G2_Gt1_F}},  //eEx_G2_Eq1,     
  {4, {eEx_G3_Eq0,  eEx_G3_Eq1,  eEx_G3_Gt1,  eEx_G3_Gt1_F}},  //eEx_G2_Gt1,   
  {4, {eEx_G3_Eq0,  eEx_G3_Eq1,  eEx_G3_Gt1,  eEx_G3_Gt1_F}},  //eEx_G2_Gt1_F, 
  //
  {4, {eEx_G3_Eq0,  eEx_G3_Eq1,  eEx_G3_Gt1,  eEx_G3_Gt1_F}},  //eEx_G3_Eq0,   
  {4, {eEx_G3_Eq0,  eEx_G3_Eq1,  eEx_G3_Gt1,  eEx_G3_Gt1_F}},  //eEx_G3_Eq1,   
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_G3_Gt1,   
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_G3_Gt1_F, 
  //    
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_Gx_Eq0,   
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_Gx_Eq1,   
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_Gx_Gt1,   
  {4, {eEx_Gx_Eq0,  eEx_Gx_Eq1,  eEx_Gx_Gt1,  eEx_Gx_Gt1_F}},  //eEx_Gx_Gt1_F, 
};



// 8x8
void get_CBP_bits_CABAC(int cbp, int b8, int* bits0, int* bits1)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int a, b;
  PixelPos block_a;
  int mb_x=(b8 & 0x01)<<1;
  int mb_y=(b8 >> 1)<<1;

  if (mb_y == 0)
  {
    if (currMB->mb_available_up == NULL)
      b = 0;
    else
    {
      if((currMB->mb_available_up)->mb_type==IPCM)
        b=0;
      else
        b = (( ((currMB->mb_available_up)->cbp & (1<<(2+(mb_x>>1)))) == 0) ? 1 : 0);   //VG-ADD
    }    
    
  }
  else
    b = ( ((cbp & (1<<(mb_x/2))) == 0) ? 1: 0);
  
  if (mb_x == 0)
  {
    getLuma4x4Neighbour(img->current_mb_nr, mb_x, mb_y, -1, 0, &block_a);
    if (block_a.available)
    {
      {
        if(img->mb_data[block_a.mb_addr].mb_type==IPCM)
          a=0;
        else
          a = (( (img->mb_data[block_a.mb_addr].cbp & (1<<(2*(block_a.y>>1)+1))) == 0) ? 1 : 0); //VG-ADD
      }
      
    }
    else
      a=0;
  }
  else
    a = ( ((cbp & (1<<mb_y)) == 0) ? 1: 0);

#ifdef	HDK_QCONTEXT 
  *bits0 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[currMB->qp*3+0] + a + 2*b, 0);
  *bits1 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[currMB->qp*3+0] + a + 2*b, 1);
#else
  *bits0 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[0] + a + 2*b, 0);
  *bits1 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[0] + a + 2*b, 1);
#endif

  return;
}

// 8x8
void estimate_CBP_bits_CABAC(int cbp, int b8, int* bits0, int* bits1, int type, int *cbp8x8_ctx)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int a, b;
  PixelPos block_a;
  int mb_x=(b8 & 0x01)<<1;
  int mb_y=(b8 >> 1)<<1;

  if (mb_y == 0)
  {
    if (currMB->mb_available_up == NULL)
      b = 0;
    else
    {
      if((currMB->mb_available_up)->mb_type==IPCM)
        b=0;
      else
        b = (( ((currMB->mb_available_up)->cbp & (1<<(2+(mb_x>>1)))) == 0) ? 1 : 0);   //VG-ADD
    }    
    
  }
  else
    b = ( ((cbp & (1<<(mb_x/2))) == 0) ? 1: 0);
  
  if (mb_x == 0)
  {
    getLuma4x4Neighbour(img->current_mb_nr, mb_x, mb_y, -1, 0, &block_a);
    if (block_a.available)
    {
      {
        if(img->mb_data[block_a.mb_addr].mb_type==IPCM)
          a=0;
        else
          a = (( (img->mb_data[block_a.mb_addr].cbp & (1<<(2*(block_a.y>>1)+1))) == 0) ? 1 : 0); //VG-ADD
      }
      
    }
    else
      a=0;
  }
  else
    a = ( ((cbp & (1<<mb_y)) == 0) ? 1: 0);

  *cbp8x8_ctx = (a+2*b);
#ifdef	HDK_QCONTEXT 
  *bits0 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[currMB->qp*3+0] + a + 2*b, 0);
  *bits1 = len_binary(img->currentSlice->tex_ctx->cbp_contexts[currMB->qp*3+0] + a + 2*b, 1);
#else
  *bits0 = len_binary(Init_ctx.cbp_contexts[type] + *cbp8x8_ctx, 0);
  *bits1 = len_binary(Init_ctx.cbp_contexts[type] + *cbp8x8_ctx, 1);
#endif

  
  return ;
}

// 4x4
void estimate_cbp_bits(int* cbp0, int* cbp1, int type, int* bcbp_ctx)
{
#define BIT_SET(x,n)  ((int)(((x)&((int64)1<<(n)))>>(n)))

  int y_ac        = (type==LUMA_16AC || type==LUMA_8x8 || type==LUMA_8x4 || type==LUMA_4x8 || type==LUMA_4x4);
  int y_dc        = (type==LUMA_16DC);
  int u_ac        = (type==CHROMA_AC && !img->is_v_block);
  int v_ac        = (type==CHROMA_AC &&  img->is_v_block);
  int chroma_dc   = (type==CHROMA_DC || type==CHROMA_DC_2x4 || type==CHROMA_DC_4x4);
  int u_dc        = (chroma_dc && !img->is_v_block);
  int v_dc        = (chroma_dc &&  img->is_v_block);
  int j           = (y_ac || u_ac || v_ac ? img->subblock_y : 0);
  int i           = (y_ac || u_ac || v_ac ? img->subblock_x : 0);
  int bit         = (y_dc ? 0 : y_ac ? 1 : u_dc ? 17 : v_dc ? 18 : u_ac ? 19 : 23);
  int default_bit = (img->is_intra_block ? 1 : 0);
  int upper_bit   = default_bit;
  int left_bit    = default_bit;
  int ctx;

  int bit_pos_a   = 0;
  int bit_pos_b   = 0;

  PixelPos block_a, block_b;

  if (y_ac || y_dc)
  {
    getLuma4x4Neighbour(img->current_mb_nr, i, j, -1,  0, &block_a);
    getLuma4x4Neighbour(img->current_mb_nr, i, j,  0, -1, &block_b);
    if (y_ac)
    {
      if (block_a.available)
        bit_pos_a = 4*block_a.y + block_a.x;
      if (block_b.available)
        bit_pos_b = 4*block_b.y + block_b.x;
    }
  }
  else
  {
    getChroma4x4Neighbour(img->current_mb_nr, i, j, -1,  0, &block_a);
    getChroma4x4Neighbour(img->current_mb_nr, i, j,  0, -1, &block_b);
    if (u_ac||v_ac)
    {
      if (block_a.available)
        bit_pos_a = 4*block_a.y + block_a.x;
      if (block_b.available)
        bit_pos_b = 4*block_b.y + block_b.x;
    }
  }

  bit = (y_dc ? 0 : y_ac ? 1 : u_dc ? 17 : v_dc ? 18 : u_ac ? 19 : 35);
  if (type!=LUMA_8x8)
  {
    if (block_b.available)
    {
      if(img->mb_data[block_b.mb_addr].mb_type==IPCM)
        upper_bit=1;
      else
        upper_bit = BIT_SET(img->mb_data[block_b.mb_addr].cbp_bits,bit+bit_pos_b);
    }

    
    if (block_a.available)
    {
      if(img->mb_data[block_a.mb_addr].mb_type==IPCM)
        left_bit=1;
      else
        left_bit = BIT_SET(img->mb_data[block_a.mb_addr].cbp_bits,bit+bit_pos_a);
    }

    ctx = 2*upper_bit+left_bit;

	*bcbp_ctx = ctx;
	*cbp0 = len_binary(Init_ctx.bcbp_contexts[type2ctx_bcbp[type]] + ctx, 0);
	*cbp1 = len_binary(Init_ctx.bcbp_contexts[type2ctx_bcbp[type]] + ctx, 1);
  }

  return;
}

void init_context(int num_coeff, int type)
{
  int i;
  TextureInfoContexts *tex_ctx = img->currentSlice->tex_ctx;

  memcpy(Init_ctx.map_contexts[type2ctx_map[type]], tex_ctx->map_contexts[type2ctx_map[type]], sizeof(BiContextType)*NUM_MAP_CTX);
  memcpy(Init_ctx.last_contexts[type2ctx_last[type]], tex_ctx->last_contexts[type2ctx_last[type]], sizeof(BiContextType)*NUM_LAST_CTX);
  memcpy(Init_ctx.one_contexts[type2ctx_one[type]], tex_ctx->one_contexts[type2ctx_one[type]], sizeof(BiContextType)*NUM_ONE_CTX);
  memcpy(Init_ctx.abs_contexts[type2ctx_abs[type]], tex_ctx->abs_contexts[type2ctx_abs[type]], sizeof(BiContextType)*NUM_ABS_CTX);

  for (i=0; i < num_coeff; i++)
  {
	memcpy(Local_ctx[i].map_contexts[type2ctx_map[type]], tex_ctx->map_contexts[type2ctx_map[type]], sizeof(BiContextType)*NUM_MAP_CTX);
	memcpy(Local_ctx[i].last_contexts[type2ctx_last[type]], tex_ctx->last_contexts[type2ctx_last[type]], sizeof(BiContextType)*NUM_LAST_CTX);
	memcpy(Local_ctx[i].one_contexts[type2ctx_one[type]], tex_ctx->one_contexts[type2ctx_one[type]], sizeof(BiContextType)*NUM_ONE_CTX);
	memcpy(Local_ctx[i].abs_contexts[type2ctx_abs[type]], tex_ctx->abs_contexts[type2ctx_abs[type]], sizeof(BiContextType)*NUM_ABS_CTX);
  }
  
  return;
}


void sdq_cabac(int* sign_coeff, int* Idx, int* scaled_coeff, double* efactor, int len, int qp_per, int* y,
			   double lambda, double* RDcost, int *num_coeff, int init_sdq_cabac_context, int type)
{
  int i, bcbp_ctx;
  tEndState End ;
  double  DiscardThr = (double) (1<<20);
  double cumdd0 = 0.0;
  int cbp0, cbp1;
  int q_bits = Q_BITS +	qp_per;
  double  discard_coef_err[64];
  Boolean state_valid[64][COEFF_STATES];
  int     quant_out[64][COEFF_STATES];
  double  quant_err[64][COEFF_STATES];
  int     coeff_rate[64][COEFF_STATES][2];
//  int ite;

  init_context(len, type);
  estimate_cbp_bits(&cbp0, &cbp1, LUMA_4x4, &bcbp_ctx);
  init_sdq_data(scaled_coeff, efactor, len, q_bits);

  for(i=len-1; i>=0; i--)
  {
    discard_coef_err[i] = cumdd0 + q_coeff[i].dist_zero;
    cumdd0 = discard_coef_err[i];
	q_coeff[i].sign = sign_coeff[i];
    
    set_valid_states(i, &state_valid[i][0] , COEFF_STATES);
  }

  End.name = eEnd;
  End.Tran[0].InStateName = eHead;
  End.Tran[0].fcost = discard_coef_err[0] + cbp0*lambda;  
#ifdef SOFTQ_RECORD_DISTORTION
  End.Tran[0].distortion = discard_coef_err[0]; //debugging only
#endif
  DiscardThr = End.Tran[0].fcost; 
  End.numT = 1;

  //for(ite=0;ite<2;ite++)
  {

	init_quant(scaled_coeff, efactor, len, q_bits, state_valid, quant_out, quant_err, coeff_rate, type);
	init_trans(C, len);
    connect_head(C, discard_coef_err, len, lambda, cbp1, state_valid, type);
    
	for(i=len-1; i>0; i--)
	{
	  update_coeff(len, C, i, DiscardThr, quant_out);
	  coeff_trans(C, i, i-1, lambda, state_valid, quant_out, quant_err, coeff_rate, type);
	}

	update_coeff(len, C, 0, DiscardThr, quant_out);
	coeff_to_end(C, len, 0, &End, lambda, quant_err, coeff_rate);
	trace_back(C, len, &End, y, quant_out);
	update_context(len, y, type);
  }	

  *RDcost = End.fminTcost / 32768.0;

  update_context_block(len, y, 0, bcbp_ctx, type);

  return;
}


void connect_head(tCoeff * C, double* discard_coef_err, int num_coeff, double lambda, int cbp1,
				  Boolean state_valid[64][COEFF_STATES], int type)
{
  int i;    
  for(i=num_coeff-1; i>=0; i--)
  {
    head_to_coeff(C, discard_coef_err, num_coeff, i, lambda, cbp1, state_valid[i], type);
  }

  return;
}

void head_to_coeff(tCoeff * C, double* discard_coef_err, int num_coeff, int i, double lambda, int cbp1,
				   Boolean state_valid[COEFF_STATES], int type)
{
  int Len;
  double dist1;
  tState *pS;
  tTransition *pT;
  tStateName sna[] = {eE0_G0_Eq1, eE0_G0_Gt1, eE0_G0_Gt1_F};

 
  for(Len=0;Len<3;Len++)
  {
    if(!state_valid[sna[Len]])
      continue;
        
    pS = &C[i].S[sna[Len]];        
    if(i==num_coeff-1)
	  dist1 = 0.0;
    else
	  dist1 = discard_coef_err[i+1];
    //qout = quant_out[i][ sna[Len] ];
	//dist2 = quant_err[i][ sna[Len] ];        
    //rate1 = coeff_rate[i][ sna[Len] ];

    pS->name = sna[Len];
    pS->bAlive = TRUE;
    pT = &pS->Tran[pS->numT];
    pS->numT++;
    pT->InCoefIndex = num_coeff;  //the head state
    pT->InStateName = eHead;
#ifdef SOFTQ_RECORD_DISTORTION
	pT->distortion = dist1;
#endif
	pT->fcost = dist1 + (cbp1)*lambda;
  }

  return;
}


void coeff_to_end(tCoeff * C, int num_coeff, int i,  tEndState * End, double lambda, double quant_err[64][COEFF_STATES], 
				  int coeff_rate[64][COEFF_STATES][2])
{
  tState *pS;
  tTransition *pT, *pT2;
  int m;
	
  for(m=0; m<COEFF_STATES; m++)
  {
	pS = &C[0].S[m];
	if(pS->bAlive)
	{
	  pT = &End->Tran[(End->numT)++];
	  pT->InCoefIndex = i;  
      pT->InStateName = pS->name;
	  
	  pT2 = &(pS->Tran[pS->minTindex]);
	  if (pT2->InStateName == eHead)
	  	 pT->fcost = pS->fminTcost + quant_err[0][m] + coeff_rate[0][m][1]*lambda;
	   else
	     pT->fcost = pS->fminTcost + quant_err[0][m] + coeff_rate[0][m][0]*lambda;
#ifdef SOFTQ_RECORD_DISTORTION
	  pT->distortion = pS->fminDistortion + quant_err[0][m];
#endif
	}
  }

  return;
}


void update_coeff(int num_coeff, tCoeff *C, int i, double DiscardThr, int quant_out[64][COEFF_STATES]) 
{
  int  m;
  for(m=0; m<COEFF_STATES; m++)
  {
	if(C[i].S[m].bAlive)
	  update_state(num_coeff, i, &C[i].S[m], DiscardThr, quant_out);
  }

  return;
}


void update_ctx_binary(BiContextTypePtr ptr_ctx, int symbol)
{
  if (symbol != ptr_ctx->MPS) 
  {
    if (! (ptr_ctx->state))
      ptr_ctx->MPS = ptr_ctx->MPS ^ 1;        // switch LPS if necessary

    ptr_ctx->state = AC_next_state_LPS_64[ptr_ctx->state]; // next state
  } 
  else 
    ptr_ctx->state = AC_next_state_MPS_64[ptr_ctx->state]; // next state

  return;
}


void Update_ctx2(BiContextTypePtr ptr_ctxpin2, int u)
{
  int k=1, exp_start=13; 
  int loop = u - 2;

  if(!loop)
  {
    update_ctx_binary(ptr_ctxpin2, 0) ;
    return;
  }
  while (((--loop)>0) && (++k <= exp_start))
  {
    update_ctx_binary(ptr_ctxpin2, 1) ; 
  }
  update_ctx_binary(ptr_ctxpin2, 1);

  if(u<15)  // u=[2,14]  ==> (u-2)x Pin2(1) +  0
  {
    //End with Pin2(0) 
    update_ctx_binary(ptr_ctxpin2, 0) ; 
  }
  else // >=15  output a 1, then  EG0
  {
    //End the current segment with Pin2(1) 
	update_ctx_binary(ptr_ctxpin2, 1) ; 
  }

  return;
}

void update_state(int num_coeff, int i, tState *pS, double DiscardThr, int quant_out[64][COEFF_STATES]) 
{
  double minT = (double) (1<<30);
#ifdef SOFTQ_RECORD_DISTORTION
  double minD = 0;
#endif
  int minTindex = 0;
  int j;
  int qo = absm(quant_out[i][pS->name]);
  tTransition *pT;
  pT = pS->Tran;


  for(j=0;j<pS->numT;j++)
  {
    if(pT->fcost < minT)
    {
      minT = pT->fcost;
#ifdef SOFTQ_RECORD_DISTORTION
	  minD = pT->distortion;
#endif
      minTindex = j;
    }
    pT++;
  }
  pS->minTindex = minTindex;
  pS->fminTcost = minT;

#ifdef SOFTQ_RECORD_DISTORTION
  pS->fminDistortion = minD;
#endif

  if(pS->fminTcost > DiscardThr)
	pS->bAlive= FALSE;

  return;
}

void coeff_trans(tCoeff *C, int iC, int jC, double lambda, Boolean state_valid[64][COEFF_STATES],
				 int quant_out[64][COEFF_STATES], double quant_err[64][COEFF_STATES], int coeff_rate[64][COEFF_STATES][2], int type)
{
  tState * pS;
  tStateName  snfrom, snto;
  int k, m;
    
  for(k=0; k<COEFF_STATES; k++) 
  {
    snfrom = (tStateName) k;
    pS = &C[iC].S[snfrom];

    if(pS->bAlive)
    {    
      for(m=0; m<TargetState[snfrom].len; m++)
      {
        snto = TargetState[snfrom].ToState[m];
         
        if(state_valid[jC][snto])
        {
          state_trans(C, iC, snfrom, jC, snto, lambda, quant_out, quant_err, coeff_rate, type);
        }
      }
    }
  }
 
  return;
}

//////////////////////////////////////////////////
//    for n=15:255
//      fx(n-14) = 2*floor( log2(n-14) ) + 1 ;
//    end
// The length of EG0 codeword for (u-15)
//////////////////////////////////////////////////
int EG0len[241] = { 1,  3,  3,  5,  5,  5,  5,  7,  7,  7,  7,  7,  7,  7,  7,
                    9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
                    9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
                   11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
                   11, 11, 11, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
                   13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
                   13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
                   13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
                   13, 13, 13, 13, 13, 13, 13, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
                   15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,  15} ;


int code_coef(BiContextTypePtr ptr_ctxpin1, BiContextTypePtr ptr_ctxpin2, int u)
{
  int loop, rate=0, k=1, exp_start=13;

  if(u==1)  
  {
    //pin1(0) 
    rate = len_binary(ptr_ctxpin1, 0);    	  
	//Pin2(none)
  }
  else  // u>1  There are 3 parts for rate2: Pin1 + Pin2 + EG0
  {
    //Pin1(1) 
    rate = len_binary(ptr_ctxpin1, 1);

    //Pin2() [u-2]x1
    loop = u-2;
	if(!loop)
	{
	  rate += len_binary(ptr_ctxpin2, 0);
	  return rate;
	}

	rate += len_binary(ptr_ctxpin2, 1);
	update_ctx_binary(ptr_ctxpin2, 1);
   
    while (((--loop)>0) && (++k <= exp_start))
	{
	  //Pin2(1)
      rate += len_binary(ptr_ctxpin2, 1);
	  //update pin2 context
	  update_ctx_binary(ptr_ctxpin2, 1); 
    }

    if(u-2<exp_start)  // u=[2,14]  ==> (u-2)x Pin2(1) +  0
    {
	  //End with Pin2(0) 
      rate += len_binary(ptr_ctxpin2, 0);
	}
	else // >=15  output a 1, then  EG0
	{
      //The tail of [15, inf)
	  rate += EG0len[u-2-exp_start] << 15;
	  //rate += est_exp_golomb_encode_eq_prob(u-exp_start-2);
	}
  }

  return rate;
}


void state_trans(tCoeff *C, int iC, tStateName snfrom, int jC, tStateName snto, double lambda,
				 int quant_out[64][COEFF_STATES], double quant_err[64][COEFF_STATES],
				 int coeff_rate[64][COEFF_STATES][2], int type)
{
  tTransition *pT, *pS2;
  tState *pSfrom = &C[iC].S[snfrom];
  tState *pSto   = &C[jC].S[snto];
  double dlev;
  int    rlev;
    
  pSto->bAlive = TRUE;
  pSto->name = snto; 
  if(pSto->bAlive)
  {	
    pT = &(pSto->Tran[(pSto->numT)++]);        
    pT->InCoefIndex = iC;
    pT->InStateName = snfrom;

    dlev = quant_err[iC][snfrom];

	pS2 = &(pSfrom->Tran[pSfrom->minTindex]);
	if (pS2->InStateName ==eHead)
		rlev = coeff_rate[iC][snfrom][1];
	else
		rlev = coeff_rate[iC][snfrom][0];

	pT->fcost = pSfrom->fminTcost + dlev + (lambda*(rlev));
#ifdef SOFTQ_RECORD_DISTORTION
	pT->distortion = pSfrom->fminDistortion + dlev;
#endif
  }

  return;
}


void init_sdq_data(int* scaled_coeff, double* efactor, int num_coeff, int q_bits)
{
  int  i, iabs, level;
//  int  zero_thresh = (1 << (q_bits - 2));   // 0.25
  int  one_thresh  = (1 << (q_bits - 2)) + (1 << (q_bits - 3));    // 0.375
  int  half_thresh = (1 << (q_bits - 1));	// 0.5

  for(i = 0; i < num_coeff; i++)
  {
	if(!scaled_coeff[i])
	{
	  q_coeff[i].num_levels = 0;
	  q_coeff[i].dist_zero = 0.0;
	  continue;
	}

	q_coeff[i].dist_zero = ((double) scaled_coeff[i]) * ((double) scaled_coeff[i]) * efactor[i];
	iabs = scaled_coeff[i] >> q_bits;
	if(iabs == 0)
	{
      q_coeff[i].num_levels = 0;

	  if(scaled_coeff[i] > one_thresh)
	  {
  	    q_coeff[i].level_low  = 1;
		level = (1 << q_bits) - scaled_coeff[i];
	    q_coeff[i].dist_low  = (double) level * (double) level * efactor[i];
	    q_coeff[i].num_levels = 1;
	  }
	}
	else
	{
  	  level = scaled_coeff[i] - (iabs << q_bits);

      if(level >= half_thresh)
	  {
		q_coeff[i].level_low = iabs;
		q_coeff[i].dist_low = (double) level * (double) level * efactor[i];

	    iabs++;
   	    level = (iabs << q_bits) - scaled_coeff[i];
		q_coeff[i].level_high = iabs;
		q_coeff[i].dist_high = (double) level * (double) level * efactor[i];
		q_coeff[i].num_levels = 2;
	  }	else if (iabs > 1)
	  {
		q_coeff[i].level_high = iabs;
		q_coeff[i].dist_high = (double) level * (double) level * efactor[i];

		iabs = 1;
		level = scaled_coeff[i] - (iabs << q_bits);
		q_coeff[i].level_low = iabs;
		q_coeff[i].dist_low = (double) level * (double) level * efactor[i];
		q_coeff[i].num_levels = 2;
	  } else
	  {
  		q_coeff[i].level_low = iabs; //iabs == 1
		q_coeff[i].dist_low = (double) level * (double) level * efactor[i];
		q_coeff[i].num_levels = 1;
	  }
	}
  }

  return;
}


void set_valid_states(int i, Boolean * bValid, int len)
{
  int k; 
  for(k=0;k<len;k++)
  {
    bValid[k] = FALSE;
  }

  if(q_coeff[i].num_levels == 0 || (q_coeff[i].num_levels == 1 && q_coeff[i].level_low == 1))
  {
    bValid[ eE1_G0_Eq0 ] = TRUE;      
    bValid[ eE2_G0_Eq0 ] = TRUE;      
	bValid[ eE3_G0_Eq0 ] = TRUE;
    bValid[ eEx_G0_Eq0 ] = TRUE;      
    bValid[ eEx_G1_Eq0 ] = TRUE;      
    bValid[ eEx_G2_Eq0 ] = TRUE;      
    bValid[ eEx_G3_Eq0 ] = TRUE;      
    bValid[ eEx_Gx_Eq0 ] = TRUE;
  }

  if(!q_coeff[i].num_levels)
    return;

  if(q_coeff[i].level_low == 1)
  {
    bValid[ eE0_G0_Eq1 ] = TRUE;      
    bValid[ eE1_G0_Eq1 ] = TRUE;      
    bValid[ eE2_G0_Eq1 ] = TRUE;      
	bValid[ eE3_G0_Eq1 ] = TRUE;      
    bValid[ eEx_G0_Eq1 ] = TRUE;      
    bValid[ eEx_G1_Eq1 ] = TRUE;      
    bValid[ eEx_G2_Eq1 ] = TRUE;      
    bValid[ eEx_G3_Eq1 ] = TRUE;      
    bValid[ eEx_Gx_Eq1 ] = TRUE; 
  }

  if(q_coeff[i].level_low > 1 || (q_coeff[i].num_levels == 2 && q_coeff[i].level_high > 1))
   {
    bValid[ eE0_G0_Gt1 ] = TRUE;      
    bValid[ eE1_G0_Gt1 ] = TRUE;      
    bValid[ eE2_G0_Gt1 ] = TRUE;      
    bValid[ eE3_G0_Gt1 ] = TRUE;      
    bValid[ eEx_G0_Gt1 ] = TRUE;      
    bValid[ eEx_G1_Gt1 ] = TRUE;      
    bValid[ eEx_G2_Gt1 ] = TRUE;      
    bValid[ eEx_G3_Gt1 ] = TRUE;      
    bValid[ eEx_Gx_Gt1 ] = TRUE;
  }

  if(q_coeff[i].num_levels == 2 && q_coeff[i].level_low > 1)
  {
    bValid[ eE0_G0_Gt1_F ] = TRUE;      
    bValid[ eE1_G0_Gt1_F ] = TRUE;      
    bValid[ eE2_G0_Gt1_F ] = TRUE;      
	bValid[ eE3_G0_Gt1_F ] = TRUE;      
    bValid[ eEx_G0_Gt1_F ] = TRUE;      
    bValid[ eEx_G1_Gt1_F ] = TRUE;      
    bValid[ eEx_G2_Gt1_F ] = TRUE;      
    bValid[ eEx_G3_Gt1_F ] = TRUE;      
    bValid[ eEx_Gx_Gt1_F ] = TRUE;  
  } 

  return;
}

void init_quant(int* scaled_coeff, double* ef, int num_coeff, int qbits,
				Boolean state_valid[64][COEFF_STATES], int quant_out[64][COEFF_STATES],
				double quant_err[64][COEFF_STATES], int coeff_rate[64][COEFF_STATES][2], int type)
{
  int  i, m, qout;
  int rate1a, rate1b, rate2;
  BiContextTypePtr ptr_ctxmap,  ptr_ctxlast;
 
  for(i=num_coeff-1; i>=0; i--)
  {
	  ptr_ctxmap  = & (Local_ctx[i].map_contexts[type2ctx_map[type]][pos2ctx_map[type][i]]);
	  ptr_ctxlast = & (Local_ctx[i].last_contexts[type2ctx_last[type]][pos2ctx_last[type][i]]);

	  for(m=eE0_G0_Eq1; m<=eEx_Gx_Gt1_F; m++)
	  {
		  coeff_rate[i][m][0] = coeff_rate[i][m][1] = 0;
		  if(state_valid[i][m])
		  {
			  if(EQUAL_ZERO(m))
			  {
				  qout = 0;
				  quant_out[i][m] = qout;
				  quant_err[i][m] = q_coeff[i].dist_zero;
				  if(i==((type == LUMA_8x8)? 63: 15))
				  {
					  rate1a = 0; //implied by cbp=1
				  } else
					  rate1a = len_binary(ptr_ctxmap, 0);
				  rate1b = rate1a;
				  rate2 = 0;
			  }
			  else 
			  {
				  if(i==((type == LUMA_8x8)? 63: 15))
				  {
					  rate1a = 0; //implied by cbp=1
					  rate1b = 0;
				  } else
				  {
					  rate1a = len_binary(ptr_ctxmap, 1) + len_binary(ptr_ctxlast, 0);
					  rate1b = len_binary(ptr_ctxmap, 1) + len_binary(ptr_ctxlast, 1);
				  }

				  if(EQUAL_ONE(m))
				  {
					  qout = q_coeff[i].sign > 0 ? 1 : -1;
					  quant_out[i][m] = qout;
					  quant_err[i][m] = q_coeff[i].dist_low;
					  rate2 = len_cabac((tStateName) m, i, qout, type);
				  }
				  else if(GT_ONE(m))
				  {
					  if(q_coeff[i].num_levels == 1)
					  {
						  qout = q_coeff[i].sign > 0 ? q_coeff[i].level_low : -q_coeff[i].level_low;
						  quant_out[i][m] = qout;
						  quant_err[i][m] = q_coeff[i].dist_low;
						  rate2 = len_cabac((tStateName) m, i, qout, type);
					  }
					  else
					  {
						  if(q_coeff[i].level_low == 1)
						  {
							  qout = q_coeff[i].sign > 0 ? q_coeff[i].level_high : -q_coeff[i].level_high;
							  quant_out[i][m] = qout;
							  quant_err[i][m] = q_coeff[i].dist_high;
							  rate2 = len_cabac((tStateName) m, i, qout, type);
						  }
						  else
						  {
							  qout = q_coeff[i].sign > 0 ? q_coeff[i].level_low : -q_coeff[i].level_low;
							  quant_out[i][m] = qout;
							  quant_err[i][m] = q_coeff[i].dist_low;
							  rate2 = len_cabac((tStateName) m, i, qout, type);
						  }
					  }
				  }
				  else //GT_ONE_FF
				  {
					  qout = q_coeff[i].sign > 0 ? q_coeff[i].level_high : -q_coeff[i].level_high;
					  quant_out[i][m] = qout;
					  quant_err[i][m] = q_coeff[i].dist_high;
					 rate2 = len_cabac((tStateName) m, i, qout, type);
				  }
				  
			  } // if EQUAL_ZERO(m)
			  coeff_rate[i][m][0]  = rate1a + rate2;
			  coeff_rate[i][m][1]  = rate1b +rate2;
		  } //if(state_valid[i][m])
	  }
  }

  return;
}


void  init_trans(tCoeff * C, int num_coeff)
{
  int i, m;

  for(i=num_coeff-1; i>=0; i--)
  {
    for(m=0; m<COEFF_STATES; m++)
    {
	  C[i].S[m].numT = 0;
      C[i].S[m].bAlive = FALSE;
    }
  }
  
  return;
}


int len_binary(BiContextTypePtr ptr_ctx, int symbol)
{
  int state = (symbol == ptr_ctx->MPS) ? 64 + ptr_ctx->state : 63 - ptr_ctx->state;
  return entropyBits[127 - state];
}


int len_cabac(tStateName m, int i, int u, int type)
{
  int rate;

  BiContextType    temp1, temp2;

  memcpy(&temp1, &Local_ctx[i].one_contexts[type2ctx_one[type]][nameCtxPin1[m]], sizeof(BiContextType));
  memcpy(&temp2, &Local_ctx[i].abs_contexts[type2ctx_abs[type]][nameCtxPin2[m]], sizeof(BiContextType));
  

  rate = code_coef(&temp1, &temp2, absm(u));
  if (absm(u))
	  rate += SIGNBIT;
  return rate;
}



//////////////////////////////////////////
//Description: update the context model for each iteration
void update_context(int num_coeff, int* y, int type)
{    
  BiContextType pin1[5], pin2[5];
  BiContextType temp_map[15], temp_last[15];
  BiContextTypePtr ptr_ctxpin1, ptr_ctxpin2;
  int i , j, p1, p2 , u, loop, k=1, exp_start=13; 
  int last;

  // Update MAP and LAST contexts  
  for (i=num_coeff-1;i>=0 && y[i]==0;i--)
  {}

  last = i;

  if (type == LUMA_8x8)
  {
	  memcpy( temp_map, &(Init_ctx.map_contexts[type2ctx_map[type]][0]), sizeof(BiContextType)*NUM_MAP_CTX);
	  memcpy( temp_last, &(Init_ctx.last_contexts[type2ctx_last[type]][0]), sizeof(BiContextType)*NUM_LAST_CTX);

	  for (i=0; i < num_coeff; i++)
	  {	  memcpy( & (Local_ctx[i].map_contexts[type2ctx_map[type]][0]), &(Init_ctx.map_contexts[type2ctx_map[type]][0]), sizeof(BiContextType)*NUM_MAP_CTX);
		  memcpy( & (Local_ctx[i].last_contexts[type2ctx_last[type]][0]), &(Init_ctx.last_contexts[type2ctx_last[type]][0]), sizeof(BiContextType)*NUM_LAST_CTX);
	  }

	  for (i = 0; i < num_coeff-1; i++)
	  {
		  memcpy( & (Local_ctx[i].map_contexts[type2ctx_map[type]][0]), temp_map, sizeof(BiContextType)*NUM_MAP_CTX);
		  memcpy( & (Local_ctx[i].last_contexts[type2ctx_last[type]][0]), temp_last, sizeof(BiContextType)*NUM_LAST_CTX);
		  ptr_ctxpin1 = &temp_map[pos2ctx_map[type][i]];
		  ptr_ctxpin2 = &temp_last[pos2ctx_last[type][i]];

		  if (y[i] != 0)
		  {	
			  update_ctx_binary(ptr_ctxpin1, 1);
			  if (i== last)
			  {
				  update_ctx_binary(ptr_ctxpin2, 1);
				  break;
			  } else
				  update_ctx_binary(ptr_ctxpin2, 0);
		  } else
		  {
			  update_ctx_binary(ptr_ctxpin1, 0);
		  }
	  }
  }


  // Update ONE and ABS contexts
  memcpy(pin1, &(Init_ctx.one_contexts[type2ctx_one[type]][0]), sizeof(BiContextType)*NUM_ONE_CTX);
  memcpy(pin2, &(Init_ctx.abs_contexts[type2ctx_abs[type]][0]), sizeof(BiContextType)*NUM_ABS_CTX);
  
  for(i=0;i<num_coeff;i++)
  {
    memcpy( &(Local_ctx[i].one_contexts[type2ctx_one[type]][0]), &(Init_ctx.one_contexts[type2ctx_one[type]][0]), sizeof(BiContextType)*NUM_ONE_CTX);
    memcpy( &(Local_ctx[i].abs_contexts[type2ctx_abs[type]][0]), &(Init_ctx.abs_contexts[type2ctx_abs[type]][0]), sizeof(BiContextType)*NUM_ABS_CTX);
  }


  p1=1;
  p2=0;
  for(j=last;j>=0;j--)
  {
	memcpy( &(Local_ctx[i].one_contexts[type2ctx_one[type]][0]), (&pin1[0]), sizeof(BiContextType)*NUM_ONE_CTX);
	memcpy( &(Local_ctx[i].abs_contexts[type2ctx_abs[type]][0]), (&pin2[0]), sizeof(BiContextType)*NUM_ABS_CTX);

	p1 = min(p1, 4);
	p2 = min(p2, 4);
	u = absm(y[j]);
	if (u)
	{
		ptr_ctxpin1 = &pin1[p1];
		ptr_ctxpin2 = &pin2[p2];

		if(u==1)  
		{
			//pin1(0) 
			update_ctx_binary(ptr_ctxpin1, 0);
			//Pin2(none)
			if (p1)
				p1++;
		}
		else if(u>1) // u>1  There are 3 parts for rate2: Pin1 + Pin2 + EG0
		{
			//Pin1(1) 
			update_ctx_binary(ptr_ctxpin1, 1);

			//Pin2() [u-2]x1
			loop = u-2;
			if(!loop)
			{
				update_ctx_binary(ptr_ctxpin2, 0);
			} else
			{
				update_ctx_binary(ptr_ctxpin2, 1);
				k = 1;
				while (((--loop)>0) && (++k <= exp_start))
				{
					update_ctx_binary(ptr_ctxpin2, 1); 
				}

				if(u-2<exp_start)  // u=[2,14]  ==> (u-2)x Pin2(1) +  0
				{
					update_ctx_binary(ptr_ctxpin2, 0);
				}			
			}
			p2++;
			p1 = 0;
		}
	}
  }

  return;
}

// 8x8: get ctx index
int get_ctx_idx(int b8)
{
  Macroblock *currMB = &img->mb_data[img->current_mb_nr];
  int a, b;
  PixelPos block_a;
  int mb_x=(b8 & 0x01)<<1;
  int mb_y=(b8 >> 1)<<1;
  int cbp = img->mb_data[img->current_mb_nr].cbp;

  if (mb_y == 0)
  {
    if (currMB->mb_available_up == NULL)
      b = 0;
    else
    {
      if((currMB->mb_available_up)->mb_type==IPCM)
        b=0;
      else
        b = (( ((currMB->mb_available_up)->cbp & (1<<(2+(mb_x>>1)))) == 0) ? 1 : 0);   //VG-ADD
    }    
    
  }
  else
    b = ( ((cbp & (1<<(mb_x/2))) == 0) ? 1: 0);
  
  if (mb_x == 0)
  {
    getLuma4x4Neighbour(img->current_mb_nr, mb_x, mb_y, -1, 0, &block_a);
    if (block_a.available)
    {
      {
        if(img->mb_data[block_a.mb_addr].mb_type==IPCM)
          a=0;
        else
          a = (( (img->mb_data[block_a.mb_addr].cbp & (1<<(2*(block_a.y>>1)+1))) == 0) ? 1 : 0); //VG-ADD
      }
      
    }
    else
      a=0;
  }
  else
    a = ( ((cbp & (1<<mb_y)) == 0) ? 1: 0);

  return a + 2*b;
}

void update_context_block(int num_coeff, int* y, int b8, int cbp_ctx, int type)
{
  int i;

  for (i=num_coeff - 1;i>=0 && y[i]==0;i--)
  {}

  if (type == LUMA_4x4)
  {
	if(!i)
		update_ctx_binary(Init_ctx.bcbp_contexts[type2ctx_bcbp[type]]+cbp_ctx, 0);
	else
		update_ctx_binary(Init_ctx.bcbp_contexts[type2ctx_bcbp[type]]+cbp_ctx, 1);
  }
  else if (type = LUMA_8x8)
  {
	  cbp_ctx = get_ctx_idx(b8);
	  if (!i)
		  update_ctx_binary(Init_ctx.cbp_contexts[0]+ cbp_ctx, 0);
	  else
		  update_ctx_binary(Init_ctx.cbp_contexts[0]+ cbp_ctx, 1);
  }

  memcpy(Init_ctx.map_contexts[type2ctx_map[type]], Local_ctx[num_coeff-1].map_contexts[type2ctx_map[type]], sizeof(BiContextType)*NUM_MAP_CTX);
  memcpy(Init_ctx.last_contexts[type2ctx_last[type]], Local_ctx[num_coeff-1].last_contexts[type2ctx_last[type]], sizeof(BiContextType)*NUM_LAST_CTX);
  memcpy(Init_ctx.one_contexts[type2ctx_one[type]], Local_ctx[num_coeff-1].one_contexts[type2ctx_one[type]], sizeof(BiContextType)*NUM_ONE_CTX);
  memcpy(Init_ctx.abs_contexts[type2ctx_abs[type]], Local_ctx[num_coeff-1].abs_contexts[type2ctx_abs[type]], sizeof(BiContextType)*NUM_ABS_CTX);

  return;
}





void trace_back(tCoeff *C, int num_coeff, tEndState *End, int *y, int quant_out[64][COEFF_STATES]) 
{
  int i;
  double minTcost = (double) (1 << 30);
#ifdef SOFTQ_RECORD_DISTORTION
  double minDcost;
#endif
  int minTindex=0;
  int CurCoef= -1;
  tTransition *pT;
  tState *pS;
  pT = End->Tran;

    
  for(i=0;i<End->numT;i++)
  {
    if(pT->fcost < minTcost)
    {
      minTcost = pT->fcost;
#ifdef SOFTQ_RECORD_DISTORTION
	  minDcost = pT->distortion;
#endif
      minTindex = i;
    }
    pT ++;
  }
  End->minTindex = minTindex;
  End->fminTcost = minTcost;
#ifdef SOFTQ_RECORD_DISTORTION
  End->fminDistortion = minDcost;
#endif

  pT = &End->Tran[minTindex];
  if(pT->InStateName == eHead)
  {
	for(i=0; i<num_coeff; i++)
	  y[i] = 0;

	return;
  }

  i = 0;
  while(pT->InStateName != eHead)
  {
    int ytmp;
    pS = &(C[pT->InCoefIndex].S[pT->InStateName]);
	ytmp= quant_out[pT->InCoefIndex][pT->InStateName];
	y[i++] = ytmp;

    CurCoef = pT->InCoefIndex;

    pT = &( pS->Tran[pS->minTindex] );

	if(pT->InStateName == eHead)
	  for( ; i<num_coeff; i++)
		y[i] = 0;
  }

  return;
}


// for 8x8
// len must be 64
void sdq_cabac8x8(int* sign_coeff, int* Idx, int* scaled_coeff, double* efactor, int len, int qp_per, int* y,
			      double lambda, double* RDcost, int *num_coeff, int init_sdq_cabac_context, int type, int b8)
{
  int i;
  tEndState End ;
  double  DiscardThr = (double) (1<<20);
  double cumdd0 = 0.0;
  int cbp0, cbp1, cbp8x8_ctx;
  int q_bits = Q_BITS_8 +	qp_per;
  double  discard_coef_err[64];
  Boolean state_valid[64][COEFF_STATES];
  int     quant_out[64][COEFF_STATES];
  double  quant_err[64][COEFF_STATES];
  int     coeff_rate[64][COEFF_STATES][2];
//  int ite;

  init_context(len, type);
  estimate_CBP_bits_CABAC(img->mb_data[img->current_mb_nr].cbp, b8, &cbp0, &cbp1, type, &cbp8x8_ctx);
  init_sdq_data(scaled_coeff, efactor, len, q_bits);

  for(i=len-1; i>=0; i--)
  {
    discard_coef_err[i] = cumdd0 + q_coeff[i].dist_zero;
    cumdd0 = discard_coef_err[i];
	q_coeff[i].sign = sign_coeff[i];
    
    set_valid_states(i, &state_valid[i][0] , COEFF_STATES);
  }

  End.name = eEnd;
  End.Tran[0].InStateName = eHead;
  End.Tran[0].fcost = discard_coef_err[0] + cbp0*lambda;  
  DiscardThr = End.Tran[0].fcost; 
  End.numT = 1;

  //for(ite=0;ite<2;ite++)
  {
	init_quant(scaled_coeff, efactor, len, q_bits, state_valid, quant_out, quant_err, coeff_rate, type);
	init_trans(C, len);
    connect_head(C, discard_coef_err, len, lambda, cbp1, state_valid, type);
    
	for(i=len-1; i>0; i--)
	{
	  update_coeff(len, C, i, DiscardThr, quant_out);
	  coeff_trans(C, i, i-1, lambda, state_valid, quant_out, quant_err, coeff_rate, type);
 	}

	update_coeff(len, C, 0, DiscardThr, quant_out);
	coeff_to_end(C, len, 0, &End, lambda, quant_err, coeff_rate);
	trace_back(C, len, &End, y, quant_out);
	update_context(len, y, type);
  }	

  *RDcost = End.fminTcost / 32768.0;

  update_context_block(len, y, b8, cbp8x8_ctx, type); 
  return;
}





