/*! ====================================================================================================================
 * \file
		v2v_enc.c
 *  \brief
		Copyright information.
 *  \par Copyright statements
		HEVC (JCTVC cfp)

		This software, including its corresponding source code, object code and executable, may only be used for 
		(1) research purposes or (2) for evaluation for standardisation purposes within the joint collaborative team on 
		video coding for HEVC , provided that this copyright notice and this corresponding notice appear in all copies, 
		and that the name of Research in Motion Limited not be used in advertising or publicity without specific, written 
		prior permission.  This software, as defined above, is provided as a proof-of-concept and for demonstration 
		purposes only; there is no representation about the suitability of this software, as defined above, for any purpose.  
 
		THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, 
		INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
		DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
		SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,PROCUREMENT OF SUBSTITUTE GOODS OR 
		SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
		WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
		USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
		
		TRADEMARKS, Product and company names mentioned herein may be the trademarks of their respective owners.  
		Any rights not expressly granted herein are reserved.
 
		Copyright (C) 2010 by Research in Motion Limited, Canada      
		All rights reserved.
 
 *  \par Full Contact Information
		Standards & Licensing Department			(standards-ipr@rim.com)
		Research in Motion
		122 West John Carpenter Parkway
		Irving, TX 75039, USA

 * ====================================================================================================================
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "global.h"
#include "v2v_enc.h"
#include "biariencode.h"

#include "trees67k.h"
#include "vlcenc.h"

parallel_buffer buffer[BUFFER_SIZE];
static uint buffer_tail;
uint vlc_pos[TREE_NUM];   /* Current node in each tree */
static uint code_pos[TREE_NUM];   /* Index of current codeword in buffer */
uint bit_pos[TREE_NUM];   /* Next free bit in current codeword */

uchar temp_space[TEMP_SIZE];
static uchar temp_offset[TEMP_SIZE];
static int code_offset;

uchar cabac_tmp_spc[TEMP_SIZE2];
uint cts_index = 0, cabac_cost = 0;

#define put_byte() { \
                     Ecodestrm[(*Ecodestrm_len)++] = Ebuffer; \
                     Ebits_to_go = 8; \
                     while (eep->C > 7) { \
                       eep->C-=8; \
                       eep->E++; \
                     } \
                    } 


void vlc_put_byte(EncodingEnvironmentPtr eep, int *Ebits_used, uchar *to) {
    assert (code_offset >= 0);
    to[*Ecodestrm_len] = code_offset;
    code_offset -= 8;
    Ecodestrm[(*Ecodestrm_len)++] = Ebuffer & 0xff;
    Ebuffer >>= 8;
    *Ebits_used -= 8;
    while (eep->C > 7) {
        eep->C-=8;
        eep->E++;
    }
} 

void v2v_init(void) {
    for (buffer_tail = 0; buffer_tail < TREE_NUM; ++buffer_tail) {
        code_pos[buffer_tail] = buffer_tail;
        bit_pos[buffer_tail] = 0;
        buffer[buffer_tail].code = buffer[buffer_tail].next = 0;
    }
    cts_index = 0;
    cabac_cost = 0;
}


extern int entropyBits[128];
void store_bit(unsigned short state, char bit) {

    if (state >= TREE_NUM) state = TREE_NUM - 1;

    if (bit_pos[state] == 32) {
        if (buffer_tail >= BUFFER_SIZE) {
            fputs("Output buffer overflow\n", stderr);
            exit(1);
        }
        buffer[buffer_tail].code = buffer[buffer_tail].next = 0;
        buffer[code_pos[state]].next = buffer_tail;
        code_pos[state] = buffer_tail++;
        bit_pos[state] = 0;
    }
    if (bit) buffer[code_pos[state]].code |= 1 << bit_pos[state];
    ++bit_pos[state];

    if (cts_index < TEMP_SIZE2) {
        cabac_tmp_spc[cts_index++] = (bit << 7) | state;
        if (cabac_cost < input->minimum_encoded_frame_size << 7 && cabac_cost < 0xff000000)
            cabac_cost += entropyBits[bit ? 64 + state : 63 - state] >> 8;
    }
}

void v2v_encode_bit(EncodingEnvironmentPtr eep, unsigned short state, uchar symbol, uchar *to) {

    int Ebits_used = 8 - Ebits_to_go;
    //* REMOVE */ fprintf(stderr, "%u\n", symbol);
    vlc_pos[state] = (enc_tree[state][vlc_pos[state]] >> 20) + !symbol;
    if ((enc_tree[state][vlc_pos[state]] >> 20) == 0) {
        int code_length = ((enc_tree[state][vlc_pos[state]] >> 16) & 0xf) + 1;
        code_offset = code_length + Ebits_used;
        Ebuffer |= (enc_tree[state][vlc_pos[state]] & 0xffff) << Ebits_used;
        Ebits_used += code_length;
        while (Ebits_used >= 8) vlc_put_byte(eep, &Ebits_used, to);
        vlc_pos[state] = 0;
    }
    Ebits_to_go = 8 - Ebits_used;
}

void vlc_encode_sym(EncodingEnvironmentPtr eep, unsigned short state, uchar symbol, uchar *to) {

    int Ebits_used = 8 - Ebits_to_go;
    uint pref_len;
    uint32 code = vlc_enc_tree[state][symbol];
    assert(0 <= Ebits_to_go && Ebits_to_go <= 9);
    assert(Ebuffer < 256);
    assert(code < 0xC0000000);

    pref_len = code & 0xf8000000;
    code_offset = (pref_len >> 27) + ((code >> 22) & 0x1f) + Ebits_used;
    if (pref_len) {
        Ebuffer |= ((1 << (pref_len >>= 27)) - 1) << Ebits_used;
        Ebits_used += pref_len;
        assert(Ebits_used <= 32);
        while (Ebits_used >= 8) vlc_put_byte(eep, &Ebits_used, to);
    }
    Ebuffer |= (code & 0x3fffff) << Ebits_used;
    Ebits_used += (code >> 22) & 0x1f;
    while (Ebits_used >= 8)
        vlc_put_byte(eep, &Ebits_used, to);
    Ebits_to_go = 8 - Ebits_used;
}

void put_pref_code(EncodingEnvironmentPtr eep, uint v) {

    assert(0 <= Ebits_to_go && Ebits_to_go <= 9);
    if (v < 128) {
        Ebuffer = v << 1;    put_byte();
        return;
    }

    v -= 128;
    if (v < 16384) {
        v = (v << 2) | 1;
        Ebuffer = v & 0xff;  put_byte();
        Ebuffer = v >> 8;    put_byte();
        return;
    }

    v -= 16384;
    if (v < 2097152) {
        v = (v << 3) | 3;
        Ebuffer = v & 0xff;  put_byte();
        v >>= 8;
        Ebuffer = v & 0xff;  put_byte();
        Ebuffer = v >> 8;    put_byte();
        return;
    }

    v -= 2097152;
    if (v >= 536870912) {
        fputs("Overflow in put_pref_code\n", stderr);
        exit(1);
    }
    v = (v << 3) | 7;
    Ebuffer = v & 0xff;      put_byte();
    v >>= 8;
    Ebuffer = v & 0xff;      put_byte();
    v >>= 8;
    Ebuffer = v & 0xff;      put_byte();
    Ebuffer = v >> 8;        put_byte();
    assert(0 <= Ebits_to_go && Ebits_to_go <= 9);
}


uint v2v_encode_seq(EncodingEnvironmentPtr eep, int tree, uchar *ts, uchar *to) {
    int k, pos = tree;
    int cl = 0;
    uint32 cw;
    unsigned char *code_buffer = Ecodestrm;
    int *code_len = Ecodestrm_len;

    Ebits_to_go = 8;
    Ebuffer = 0;
    Ecodestrm = ts;
    Ecodestrm_len = &cl;

    while (buffer[pos].next) {
        cw = buffer[pos].code;
        for (k = 0; k < 32; ++k, cw >>= 1)
            v2v_encode_bit(eep, tree, cw & 1, to);
        pos = buffer[pos].next;
    }

    cw = buffer[pos].code;
    for (k = 0; k < bit_pos[tree]; ++k, cw >>= 1)
        v2v_encode_bit(eep, tree, cw & 1, to);

    //v2v_flush_if_needed(eep, tree);
    if (vlc_pos[tree] && (enc_tree[tree][vlc_pos[tree]] >> 20)) {
        int Ebits_used = 8 - Ebits_to_go;
        int code_length;
        do {
            vlc_pos[tree] = (enc_tree[tree][vlc_pos[tree]] >> 20) + 1;
        } while (enc_tree[tree][vlc_pos[tree]] >> 20);
        code_length = ((enc_tree[tree][vlc_pos[tree]] >> 16) & 0xf) + 1;
        code_offset = code_length + Ebits_used;
        Ebuffer |= (enc_tree[tree][vlc_pos[tree]] & 0xffff) << Ebits_used;
        Ebits_used += code_length;
        while (Ebits_used >= 8) vlc_put_byte(eep, &Ebits_used, to);
        vlc_pos[tree] = 0;
        Ebits_to_go = 8 - Ebits_used;
    }
    if (Ebits_to_go != 8) { put_byte(); }
    //else Ecodestrm[(*Ecodestrm_len)++] = 0;

    Ecodestrm = code_buffer;
    Ecodestrm_len = code_len;
    return cl;
}

uint vlc_encode_seq(EncodingEnvironmentPtr eep, int tree, uchar *ts, uchar *to) {
    int k, pos = tree;
    int cl = 0;
    uint32 cw;
    unsigned char *code_buffer = Ecodestrm;
    int *code_len = Ecodestrm_len;

    Ebits_to_go = 8;
    Ebuffer = 0;
    Ecodestrm = ts;
    Ecodestrm_len = &cl;

    while (buffer[pos].next) {
        cw = buffer[pos].code;
        vlc_encode_sym(eep, tree, cw & 0xff, to);  cw >>= 8;
        vlc_encode_sym(eep, tree, cw & 0xff, to);  cw >>= 8;
        vlc_encode_sym(eep, tree, cw & 0xff, to);  cw >>= 8;
        vlc_encode_sym(eep, tree, cw & 0xff, to);
        pos = buffer[pos].next;
    }

    cw = buffer[pos].code;
    for (k = bit_pos[tree]; k > 0; k -= 8, cw >>= 8)
        vlc_encode_sym(eep, tree, cw & 0xff, to);

    if (Ebits_to_go != 8) put_byte();

    Ecodestrm = code_buffer;
    Ecodestrm_len = code_len;
    return cl;
}

void vlc_done (EncodingEnvironmentPtr eep, int *offset) {
    int k;
    uint total_len = 0;

    for (k = 0; k < TREE_NUM; ++k) {
        //* REMOVE */ fprintf(stderr,"\nSEQ %u\n", k);
        uint len = active_pps->binary_coding_mode == Variable ?
                      v2v_encode_seq(eep, k, temp_space + *offset, temp_offset + *offset) :
                      vlc_encode_seq(eep, k, temp_space + *offset, temp_offset + *offset);
        put_pref_code(eep, len);
        *offset += len;
        total_len += len;
    }
    temp_space[(*offset)++] = 0x81;

    for (k = 1; k < active_pps->balanced_cpus; ++k) {
        put_pref_code(eep, temp_offset[k * total_len / active_pps->balanced_cpus]);
    }
}


